#import <Foundation/NSString.h>
#import <Foundation/NSValue.h>
#import <Foundation/NSData.h>
#import <Foundation/NSDate.h>
#import <Foundation/NSArray.h>
#import <Foundation/NSDictionary.h>
#import <Foundation/NSEnumerator.h>
#import <Foundation/NSNull.h>
#import <Foundation/NSAutoreleasePool.h>
#import <AppKit/NSPanel.h>
#import <AppKit/NSPasteboard.h>
#import <libkern/OSTypes.h>
#import "EXTERN.h"
#import "perl.h"
#import "XSUB.h"
#import "ppport.h"
#import "bh_pboard.h"
#import "bh_panel.h"
#import "bh_finder.h"

// ===================================================================
// PERL BRIDGE CODE
// ===================================================================


// ----------------------------------------------------------------------
// OBJECT TYPING -- All of these methods assume the script object is a
// scalar value. For arrays and hashes, you must use references. 

typedef SV * BHScriptObject;

typedef enum {
    BHUnknownType = 0,
    BHStringType = 4,
        BHUnicodeStringType = 5,
    BHNumberType = 8,
        BHIntegerType = 9,
        BHLongType = 11,
        BHFloatType = 13,
    BHBinaryType = 16,
    BHTimestampType = 32,
        BHTimeOnlyType = 33,
    BHArrayType = 64,
    BHDictionaryType = 128
} BHObjectType;


BHObjectType BHTypeForNativeObject(id cfObj)
{
    if (NULL == cfObj) { return BHUnknownType; }
 
	if ([cfObj isKindOfClass:[NSString class]]) { return BHStringType; }
	if ([cfObj isKindOfClass:[NSNumber class]]) { return BHNumberType; }
	if ([cfObj isKindOfClass:[NSData class]]) { return BHBinaryType; }
	if ([cfObj isKindOfClass:[NSDate class]]) { return BHTimestampType; }
	if ([cfObj isKindOfClass:[NSArray class]]) { return BHArrayType; }
	if ([cfObj isKindOfClass:[NSDictionary class]]) { return BHDictionaryType; }

    return BHUnknownType;
}

BHObjectType BHTypeForScriptObject(BHScriptObject plObj)
{
	if (NULL == plObj) { return BHUnknownType; }
	// check for unknown!!
	
	if (SvROK(plObj)) {
		U32 plType = SvTYPE(SvRV(plObj));

		if (SVt_PV == plType) { return BHStringType; }
		if (SVt_IV == plType) { return BHIntegerType; }
		if (SVt_NV == plType) { return BHFloatType; }
		if (SVt_PVAV == plType) { return BHArrayType; }
		if (SVt_PVHV == plType) { return BHDictionaryType; }
	}
	else {
		if (SvIOK(plObj)) { return BHIntegerType; }
		if (SvNOK(plObj)) { return BHFloatType; }
		if (SvPOK(plObj)) { return BHStringType; }
	}
	
    return BHUnknownType;    
}

int BHIsNull(BHScriptObject plObj)
{
	if (NULL == plObj) { return 1; }
	
	BHObjectType plType = BHTypeForScriptObject(plObj);
	SV *sv = SvROK(plObj) ? SvRV(plObj) : plObj;
	
	switch (plType) {
	case BHIntegerType:
	case BHStringType:
	case BHFloatType:
		return (SvTRUE(sv) ? 0 : 1);
	case BHArrayType:
	case BHDictionaryType:
		return 0;
	default:
		return 1;
	}
}

id BHConvertToNativeObject(BHScriptObject plObj);
BHScriptObject BHConvertToScriptObject(id cfObj);


// ----------------------------------------------------------------------
// RETURN VALUE CONVENIENCES

BHScriptObject BHScriptNullObject()
{
    return &PL_sv_undef;
}

BHScriptObject BHScriptIntObject(int i)
{
	return newSViv(i);
}

BHScriptObject BHScriptBoolObject(int flag)
{
	if (0 != flag) { 
		return BHScriptIntObject(1);
	}
	
	return BHScriptIntObject(0);
}


// ----------------------------------------------------------------------
// OBJECT CONVERSION

id BHConvertToNativeString(BHScriptObject plStr)
{
	SV *sv;
	
    if (BHIsNull(plStr)) { return nil; }

	if (SvROK(plStr)) {
		sv = SvRV(plStr);
	}
	else {
		sv = plStr;
	}
	
    STRLEN numbytes;
	UInt8 *bytes = (UInt8 *)SvPV(sv, numbytes);
	id s = [[NSString alloc] initWithBytes:bytes length:numbytes encoding:NSUTF8StringEncoding];
	return [s autorelease];
}

BHScriptObject BHConvertToScriptString(id cfObj)
{
	id d = [cfObj dataUsingEncoding:NSUTF8StringEncoding];
	int l = [d length];
	SV *sv = newSVpvn((char *)[d bytes], l);
	return sv;
}

id BHConvertToNativeNumber(BHScriptObject plNum)
{
	SV *sv;
	
    if (BHIsNull(plNum)) { return NULL; }

	if (SvROK(plNum)) {
		sv = SvRV(plNum);
	}
	else {
		sv = plNum;
	}
	

	id cfNum;
	if (SvIOK(sv)) {
		int i = SvIV(sv);
		cfNum = [NSNumber numberWithInt:i];
	}
	else {
		double d = SvNV(sv);
		cfNum = [NSNumber numberWithDouble:d];
	}
	
	return cfNum;
}

BHScriptObject BHConvertToScriptNumber(id cfNum)
{
	SV *sv;
	
	switch ((int)[cfNum objCType]) {
		case 'f':
		case 'd':
			sv = newSVnv([cfNum doubleValue]);
			break;
		default:
			sv = newSViv([cfNum intValue]);
	}
	
	return sv;
}

id BHConvertToNativeData(BHScriptObject plData)
{
	SV *sv;
	
    if (BHIsNull(plData)) { return nil; }
	
	if (SvROK(plData)) {
		sv = SvRV(plData);
	}
	else {
		sv = plData;
	}
	
    STRLEN numbytes;
	UInt8 *bytes = (UInt8 *)SvPV(sv, numbytes);
	NSData *d = [NSData dataWithBytes:bytes length:numbytes];
	return d;
}

BHScriptObject BHConvertToScriptData(id cfData)
{
	int len = [cfData length];
	SV *sv = newSVpvn((char *)[cfData bytes], len);
	return sv;
}

id BHConvertToNativeArray(BHScriptObject plArray)
{
	if (BHIsNull(plArray)) { return NULL; }
	
	AV *av = (AV *)SvRV(plArray);
	
	int len = av_len(av) + 1;
	int i;
	id cfArray = [NSMutableArray array]; 
	
	for (i=0; i<len; i++) {
		SV **elem;
		elem = av_fetch(av,i,0);
		id aRef = BHConvertToNativeObject(*elem);
		[cfArray addObject:aRef];
	}
	
	return cfArray;
}

BHScriptObject BHConvertToScriptArray(id cfArray)
{
	int num = [cfArray count];
	I32 i;
	AV *av = newAV();
	
	for (i=0; i<num; i++) {
		id elem = [cfArray objectAtIndex:i];
		BHScriptObject sv = BHConvertToScriptObject(elem);
		av_push(av,sv);
	}
	
	return newRV((SV*)av);
}

id BHConvertToNativeDictionary(BHScriptObject plDict)
{
	if (BHIsNull(plDict)) { return NULL; }
	
	id d = [NSMutableDictionary dictionary];
	HV *hv = (HV *)SvRV(plDict);
	I32 num = hv_iterinit(hv);
	int i;
	
	for (i=0; i<num; i++) {
		HE *he = hv_iternext(hv);
		SV *plKey = hv_iterkeysv(he);
		SV *plVal = hv_iterval(hv, he);
		id key = BHConvertToNativeString(plKey);
		id val = BHConvertToNativeObject(plVal);
		if (nil != key) {
			if (nil == val) { val = [NSNull null]; }
			[d setObject:val forKey:key];
		}
	}
	
	return d;
}

BHScriptObject BHConvertToScriptDictionary(id cfDict)
{
	HV *hv = newHV();
	id enumer = [cfDict keyEnumerator];
	id aKey;
	
	while (aKey = [enumer nextObject]) {
		id aVal = [cfDict objectForKey:aKey];
		SV *hKey = BHConvertToScriptString(aKey);
		SV *hVal = BHConvertToScriptObject(aVal);
		HE *hEnt = hv_store_ent(hv, hKey, hVal, 0);
	}
	
	return newRV((SV*)hv);
}


/* Recursively  converts plObj, which is a perl object, into a core foundation type ref. For container classes, this function will visit all of the child objects in the container, and call the function recursively for those children. This way, a deep conversion is made of the property list container classes (e.g. NSArray). */
id BHConvertToNativeObject(BHScriptObject plObj)
{
    BHObjectType objType = BHTypeForScriptObject(plObj);
    
    if (BHStringType & objType)     { return BHConvertToNativeString(plObj); }
    if (BHNumberType & objType)     { return BHConvertToNativeNumber(plObj); }
    if (BHArrayType & objType)      { return BHConvertToNativeArray(plObj); }
    if (BHDictionaryType & objType) { return BHConvertToNativeDictionary(plObj); }
    
    return NULL;
}

/* Recursively  converts cfObj, which is a core foundation type ref, into a perl object. For container classes, this function will visit all of the child objects in the container, and call the function recursively for those children. This way, a deep conversion is made of the property list container classes (e.g. NSArray). */
BHScriptObject BHConvertToScriptObject(id cfObj) 
{
    BHObjectType objType = BHTypeForNativeObject(cfObj);
    
    if (BHStringType & objType)     { return BHConvertToScriptString(cfObj); }
    if (BHNumberType & objType)     { return BHConvertToScriptNumber(cfObj); }
    if (BHBinaryType & objType)     { return BHConvertToScriptData(cfObj); }
    if (BHArrayType & objType)      { return BHConvertToScriptArray(cfObj); }
    if (BHDictionaryType & objType) { return BHConvertToScriptDictionary(cfObj); }
    
    return BHScriptNullObject();
}


// ===================================================================
// CONSTANTS BRIDGING
// ===================================================================
#include "bhglobals.xsh"


// ===================================================================
// FUNCTIONAL MODULES
// ===================================================================
MODULE = PLext		PACKAGE = PLext

BOOT:
add_cocoa_globals();


# --------------------------------------------------------------------
# PASTEBOARD, PANEL & FINDER
#

INCLUDE:	cat bhpboard.xsh |
INCLUDE:	cat bhpanel.xsh |
INCLUDE:	cat bhfinder.xsh |
